
#include "../gfx/opengl.hpp"
#include "Window.hpp"
#include "System.hpp"

using namespace std;

namespace bleifrei {

	platform::Window Singleton<platform::Window>::instance;

	namespace platform {

		const char *CLASS_NAME = "BleifreiWinClass";

		void
		Window::toLocal(int &x, int &y)
		{
			POINT p = { x, y };
			ScreenToClient(windowHandle, &p);
			x = p.x;
			y = p.y;
		}

		void
		Window::setEventListener(WindowEventListener *eventListener)
		{
			this->eventListener = eventListener;
			if (eventListener) {
				eventListener->onResize(width, height);
			}
		}

		bool
		Window::init(const string &title, int width, int height, int bpp, bool fullscreen)
		{
			WNDCLASSEX windowClass;
			DWORD extendedWindowStyle;
			DWORD windowStyle;
			RECT windowRect;

			this->title = title;
			this->width = width;
			this->height = height;
			this->bpp = bpp;
			this->fullscreen = fullscreen;
			doubleBuffered = true;

			HINSTANCE instanceHandle = System::getInstance().windows_getInstanceHandle();

			if (fullscreen) this->fullscreen = windows_fullscreen();

			windowRect.top = 0;
			windowRect.left = 0;
			windowRect.bottom = height;
			windowRect.right = width;
			
			windowClass.cbSize			= sizeof(WNDCLASSEX);	
			windowClass.style			= CS_HREDRAW | CS_VREDRAW;
			windowClass.lpfnWndProc		= System::windows_callback;
			windowClass.cbClsExtra		= 0;
			windowClass.cbWndExtra		= 0;
			windowClass.hInstance		= instanceHandle;
			windowClass.hIcon			= LoadIcon(NULL, IDI_APPLICATION);
			windowClass.hCursor			= LoadCursor(NULL, IDC_ARROW);
			windowClass.hbrBackground	= NULL;
			windowClass.lpszMenuName	= NULL;
			windowClass.lpszClassName	= CLASS_NAME;
			windowClass.hIconSm			= LoadIcon(NULL,IDI_APPLICATION);

			if (!RegisterClassEx(&windowClass)) return false;

			if (fullscreen)	{
				extendedWindowStyle =	WS_EX_APPWINDOW;
				windowStyle = WS_POPUP;
				ShowCursor(false);
			}
			else {
				extendedWindowStyle = NULL;
				windowStyle =	WS_OVERLAPPEDWINDOW |
								WS_VISIBLE |
								WS_SYSMENU |
								WS_CLIPCHILDREN |
								WS_CLIPSIBLINGS;

				AdjustWindowRectEx(&windowRect, windowStyle, false, extendedWindowStyle);
				windowRect.bottom -= windowRect.top;
				windowRect.right -= windowRect.left;
				windowRect.left = 0;
				windowRect.top = 0;
			}

			windowHandle = CreateWindowEx(extendedWindowStyle,
							CLASS_NAME,
							title.c_str(),
							windowStyle,
							windowRect.left, windowRect.top,
							windowRect.right, windowRect.bottom,
							NULL,
							NULL,
							instanceHandle,
							NULL);

			if (!windowHandle) return false;
			
			ShowWindow(windowHandle, SW_SHOW);
			UpdateWindow(windowHandle);

			hDC = GetDC(windowHandle);
			windows_pixelFormat();

			hRC = wglCreateContext(hDC);
			wglMakeCurrent(hDC, hRC);

			SetWindowLongPtr(windowHandle, GWLP_USERDATA, (LONG)(LONG_PTR)this);

			if (fullscreen)
			{
				m_startMenu = FindWindow("Shell_TrayWnd", "");
				SetWindowPos(m_startMenu, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
				SetWindowPos(windowHandle, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
			}

			return true;
		}

		void
		Window::destroy(void)
		{
			wglMakeCurrent(hDC, NULL);
			wglDeleteContext(hRC);
			if (fullscreen) ChangeDisplaySettings(NULL, 0);

			DestroyWindow(windowHandle);
			UnregisterClass(CLASS_NAME, System::getInstance().windows_getInstanceHandle());
		}

		bool
		Window::windows_fullscreen(void)
		{
			DEVMODE devModeScreen;
			memset(&devModeScreen, 0, sizeof(devModeScreen));
			devModeScreen.dmSize = sizeof(devModeScreen);
			devModeScreen.dmPelsWidth = width;
			devModeScreen.dmPelsHeight = height;
			devModeScreen.dmBitsPerPel = bpp;
			devModeScreen.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;

			return (ChangeDisplaySettings(&devModeScreen, CDS_FULLSCREEN) == DISP_CHANGE_SUCCESSFUL);
		}

		void
		Window::windows_pixelFormat(void)
		{
			int nPixelFormat;

			PIXELFORMATDESCRIPTOR pfd = {
					sizeof(PIXELFORMATDESCRIPTOR),
					1,
					PFD_DRAW_TO_WINDOW |
					PFD_SUPPORT_OPENGL |
					PFD_DOUBLEBUFFER,
					PFD_TYPE_RGBA,
					bpp,
					0,0,0,0,0,0,
					0,
					0,
					0,
					0,0,0,0,
					32,
					8,
					0,
					PFD_MAIN_PLANE,
					0,
					0,0,0 };

			nPixelFormat = ChoosePixelFormat(hDC, &pfd);

			SetPixelFormat(hDC, nPixelFormat, &pfd);
		}

		void
		WindowEventListener::onClose(void)
		{
		}

		void
		WindowEventListener::onResize(int width, int height)
		{
		}

	}
}