
// includes
#include <iostream>

#include "../gfx/Effect.hpp"
#include "../io/io.hpp"
#include "../tex/Texture.hpp"
#include "../tex/TextureManager.hpp"
#include "../Base.hpp"
#include "Material.hpp"

using namespace bleifrei::io;
using namespace bleifrei::gfx;
using namespace bleifrei::math;
using namespace bleifrei::tex;
using namespace std;

namespace bleifrei {
	namespace scene {

		Material::Material(void) :
			m_diffuseMap1(0),
			m_diffuseMap2(0),
			m_opacityMap(0),
			m_reflectionMap(0),
			m_bumpMap(0),
			m_specularMap(0),
			m_ambient(Vector3f(0,0,0)),
			m_diffuse(Vector3f(0,0,0)),
			m_specular(Vector3f(0,0,0)),
			m_shininess(0),
			m_transparency(0)
		{
		}

		Material::Material(const Material &material) :
			m_diffuseMap1(0),
			m_diffuseMap2(0),
			m_opacityMap(0),
			m_reflectionMap(0),
			m_bumpMap(0),
			m_specularMap(0),
			m_ambient(material.m_ambient),
			m_diffuse(material.m_diffuse),
			m_specular(material.m_specular),
			m_shininess(material.m_shininess),
			m_transparency(material.m_transparency)
		{
			this->setDiffuseMap1(material.m_diffuseMap1);
			this->setDiffuseMap2(material.m_diffuseMap2);
			this->setOpacityMap(material.m_opacityMap);
			this->setReflectionMap(material.m_reflectionMap);
			this->setBumpMap(material.m_bumpMap);
			this->setSpecularMap(material.m_specularMap);
		}

		Material::~Material(void)
		{
			safe_release(m_diffuseMap1);
			safe_release(m_diffuseMap2);
			safe_release(m_opacityMap);
			safe_release(m_reflectionMap);
			safe_release(m_bumpMap);
			safe_release(m_specularMap);
		}

		void
		Material::setDiffuseMap1(tex::Texture *diffuseMap)
		{
			safe_release(m_diffuseMap1);
			m_diffuseMap1 = diffuseMap;
			
			if (m_diffuseMap1) m_diffuseMap1->addReference();
		}

		void
		Material::setDiffuseMap2(tex::Texture *diffuseMap)
		{
			safe_release(m_diffuseMap2);
			m_diffuseMap2 = diffuseMap;
			
			if (m_diffuseMap2) m_diffuseMap2->addReference();
		}

		void
		Material::setOpacityMap(tex::Texture *opacityMap)
		{
			safe_release(m_opacityMap);
			m_opacityMap = opacityMap;
			
			if (m_opacityMap) m_opacityMap->addReference();
		}

		void
		Material::setReflectionMap(tex::Texture *reflectionMap)
		{
			safe_release(m_reflectionMap);
			m_reflectionMap = reflectionMap;
			
			if (m_reflectionMap) m_reflectionMap->addReference();
		}

		void
		Material::setBumpMap(tex::Texture *bumpMap)
		{
			safe_release(m_bumpMap);
			m_bumpMap = bumpMap;
			
			if (m_bumpMap) m_bumpMap->addReference();
		}

		void
		Material::setSpecularMap(tex::Texture *specularMap)
		{
			safe_release(m_specularMap);
			m_specularMap = specularMap;
			
			if (m_specularMap) m_specularMap->addReference();
		}

		void
		Material::read(istream &is)
		{
			FileId id;
			get(is, id);

			if (!is)
			{
				// $TODO: signal error "Failed to load file id!"
				return;
			}

			if (id != string2id("MTRL"))
			{
				// $TODO: signal error "Not a valid mesh2 file!"
				return;
			}

			FileVersion ver;
			get(is, ver);

			if (ver.major != 1 || ver.minor != 0)
			{
				// $TODO: signal error "Unknown mesh version!"
				return;
			}

			get(is, m_name);

			TextureManager &textureManager = TextureManager::getInstance();

			string textureName;

			get(is, textureName);
			setDiffuseMap1(textureManager.load(textureName));
			get(is, textureName);
			setDiffuseMap2(textureManager.load(textureName));
			get(is, textureName);
			setOpacityMap(textureManager.load(textureName));
			get(is, textureName);
			setReflectionMap(textureManager.load(textureName));
			get(is, textureName);
			setBumpMap(textureManager.load(textureName));
			get(is, textureName);
			setSpecularMap(textureManager.load(textureName));

			get(is, m_ambient);
			get(is, m_diffuse);
			get(is, m_specular);

			get(is, m_shininess);
			get(is, m_transparency);
		}

		void
		Material::write(ostream &os)
		{
			FileId id = string2id("MTRL");
			put(os, id);

			FileVersion ver;
			ver.major =	1;
			ver.minor = 0;
			put(os, ver);

			put(os, m_name);

			TextureManager &textureManager = TextureManager::getInstance();

			put(os, textureManager.getTextureName(m_diffuseMap1));
			put(os, textureManager.getTextureName(m_diffuseMap2));
			put(os, textureManager.getTextureName(m_opacityMap));
			put(os, textureManager.getTextureName(m_reflectionMap));
			put(os, textureManager.getTextureName(m_bumpMap));
			put(os, textureManager.getTextureName(m_specularMap));

			put(os, m_ambient);
			put(os, m_diffuse);
			put(os, m_specular);

			put(os, m_shininess);
			put(os, m_transparency);
		}

	}
}