/** 
 * bleifrei utility library
 *
 * Copyright (C) 2003 Ulf Borgenstam, Jonas Svensson.
 *
 * Do not distribute. If you know someone interested in obtaining these sources
 * please have them contact us.
 */

// includes
#include "../io/console.hpp"
#include "../io/io.hpp"
#include "../math/Matrix.hpp"
#include "../math/Matrix_op.hpp"
#include "../platform/types.hpp"

#include "Mesh.hpp"

using namespace bleifrei::io;
using namespace bleifrei::math;
using namespace bleifrei::platform;
using namespace std;

namespace bleifrei {
	namespace scene {

		#pragma pack(push, 1)

		struct MeshHeader
		{
			dword vertices;
			dword triangles;
		};

		#pragma pack(pop)


		Mesh::Mesh(void)
		{
		}

		Mesh::~Mesh(void)
		{
		}

		void
		Mesh::append(Mesh *mesh)
		{
			unsigned int size = (unsigned int)vertices.size();
			Vector3ui offset = Vector3ui(size, size, size);

			vertices.insert(vertices.end(), mesh->vertices.begin(), mesh->vertices.end());
			normals.insert(normals.end(), mesh->normals.begin(), mesh->normals.end());
			texcoords.insert(texcoords.end(), mesh->texcoords.begin(), mesh->texcoords.end());

			materials.insert(materials.end(), mesh->materials.begin(), mesh->materials.end());

			vector<Vector3ui>::iterator tit;
			for (tit = mesh->triangles.begin(); tit != mesh->triangles.end(); ++tit)
			{
				triangles.push_back(*tit + offset);
			}
		}

		unsigned int
		Mesh::addVertex(const Vector3f &vertex) 
		{
			vertices.push_back(vertex);
			texcoords.push_back(Vector2f(0,0));
			normals.push_back(Vector3f(0,1,0));
			return (unsigned int)(vertices.size() - 1);
		}

		unsigned int
		Mesh::addTriangle(const Vector3ui &triangleIndices, unsigned int materialIndex)
		{
			triangles.push_back(triangleIndices);
			materials.push_back(materialIndex);
			return (unsigned int)(triangles.size() - 1);
		}

		void
		Mesh::setVertex(unsigned int index, const Vector3f &vertex)
		{
			vertices[index] = vertex;
		}

		void 
		Mesh::setNormal(unsigned int index, const Vector3f &normal)
		{
			normals[index] = normal;
		}

		void
		Mesh::setTexCoord(unsigned int index, const Vector2f &texCoord)
		{
			texcoords[index] = texCoord;
		}

		void
		Mesh::read(istream &is)
		{
			FileId id;
			get(is, id);

			if (!is)
			{
				// $TODO: signal error "Failed to load file id!"
				return;
			}

			if (id != string2id("mSH2"))
			{
				// $TODO: signal error "Not a valid mesh2 file!"
				return;
			}

			FileVersion ver;
			get(is, ver);

			if (ver.major != 1 || ver.minor != 0)
			{
				// $TODO: signal error "Unknown mesh version!"
				return;
			}

			MeshHeader header;

			get(is, header);									
			get(is, name);

			for (unsigned int i = 0; i < header.vertices; i++)
			{
				Vector3f vertex;
				get(is, vertex);
				addVertex(vertex);
			}
			
			for (unsigned int i = 0; i < header.vertices; i++)
			{
				Vector3f normal;
				get(is, normal);
				setNormal(i, normal);
			}
			
			for (unsigned int i = 0; i < header.vertices; i++)
			{
				Vector2f texCoord;
				get(is, texCoord);
				setTexCoord(i, texCoord);
			}
			
			for (unsigned int i = 0; i < header.triangles; i++)
			{
				Vector4ui triangle;
				get(is, triangle);
				addTriangle(Vector3ui(&triangle[0]), triangle[3]);
			}
		}

		void
		Mesh::write(ostream &os)
		{
			FileId id = string2id("mSH2");
			put(os, id);

			FileVersion ver;
			ver.major =	1;
			ver.minor = 0;
			put(os, ver);

			MeshHeader header;
			header.vertices = (dword)vertices.size();
			header.triangles = (dword)triangles.size();

			put(os, header);
			put(os, name);

			for (vector<Vector3f>::iterator it = vertices.begin(); it != vertices.end(); ++it)
			{
				put(os, *it);
			}

			for (vector<Vector3f>::iterator it = normals.begin(); it != normals.end(); ++it)
			{
				put(os, *it);
			}

			for (vector<Vector2f>::iterator it = texcoords.begin(); it != texcoords.end(); ++it)
			{
				put(os, *it);
			}

			vector<unsigned int>::iterator itMaterial = materials.begin();
			for (vector<Vector3ui>::iterator it = triangles.begin(); it != triangles.end(); ++it, ++itMaterial)
			{
				put(os, Vector4ui((*it)[0], (*it)[1], (*it)[2], *itMaterial));
			}
		}


	}
}